const std = @import("std");
const mem = std.mem;
const Terminal = @import("terminal.zig");
const Cell = Terminal.Cell;
const Buffer = Terminal.Buffer;
const Color = @import("color.zig").RGBA;
const Control = @This();

allocator: mem.Allocator,

parent: ?*Control,
children: std.ArrayList(*Control),

rect: ?Rect = null, // TODO: maybe unionize this with .parent

left: u32,
right: u32,
top: u32,
bottom: u32,

border_type: enum { line, bg, none } = .none,
border_char: u8 = ' ',
border_bg: Color = Color{ .r = 0.0, .g = 0.0, .b = 0.0 },
border_fg: Color = Color{ .r = 1.0, .g = 1.0, .b = 1.0 },

content: std.ArrayList([]const u8),
content_halign: enum { left, center, right } = .left,
content_valign: enum { top, center, bottom } = .top,

_draw_buf: ?*Buffer = null,

pub fn init(allocator: mem.Allocator) Control {
    return .{
        .allocator = allocator,
        .parent = null,
        .children = std.ArrayList(*Control).init(allocator),
        .left = 2,
        .right = 2,
        .top = 1,
        .bottom = 1,
        .content = std.ArrayList([]const u8).init(allocator),
    };
}

pub fn deinit(self: *Control) void {
    if (self._draw_buf) |b| b.deinit();
    for (self.children.items) |child| child.deinit();
    self.children.deinit();
    self.content.deinit();
}

pub fn addChild(self: *Control, child: *Control) !void {
    std.debug.assert(child.parent == null);
    child.parent = self;
    try self.children.append(child);
}

pub fn removeChild(self: *Control, child: *Control) !void {
    _ = self;
    _ = child;
}

pub fn removeChildByIndex(self: *Control, child: usize) !void {
    _ = self;
    _ = child;
}

pub fn moveChild(self: *Control, child: *Control, idx: usize) !void {
    _ = self;
    _ = child;
    _ = idx;
}

pub fn moveChildByIndex(self: *Control, child: usize, idx: usize) !void {
    _ = self;
    _ = child;
    _ = idx;
}

pub fn makeDirty(self: *Control) void {
    if (self._draw_buf) |b| b.deinit();
    self._draw_buf = null;
    for (self.children.items) |child| child.makeDirty();
}

pub fn draw(self: *Control, term: *Terminal) !void {
    if (self._draw_buf) |draw_buf|
        try draw_buf.draw(term)
    else {
        const rect = self.getRect();
        var buffer = try Buffer.init(self.allocator, rect.h, rect.w);
        errdefer buffer.deinit();

        switch (self.border_type) {
            .none => {
                var y: u32 = 0;
                // try term.cursorSet(rect.x, rect.y);
                // while (y < rect.h) : (y += 1) {
                //     try term.cursorSet(rect.x, rect.y + y);
                //     var x: u32 = 0;
                //     while (x < rect.w) : (x += 1) try term.print(" ", .{});
                // }
                if (self.content.items.len > 0) {
                    var lines = self.content.items.len;
                    for (self.content.items) |line| lines += mem.count(u8, line, "\n");
                    const tx = switch (self.content_halign) {
                        .left => rect.x,
                        .center => rect.x + (rect.w / 2) - (@as(u32, @intCast(self.content.items.len)) / 2),
                        .right => rect.x + rect.w - (@as(u32, @intCast(self.content.items.len))),
                    };
                    var x = tx;
                    y = switch (self.content_valign) {
                        .top => rect.y,
                        .center => rect.y + (rect.h / 2) - (@as(u32, @intCast(lines)) / 2),
                        .bottom => rect.y + rect.h - 1 - @as(u32, @intCast(lines)),
                    };

                    for (self.content.items) |line| {
                        var split = mem.splitScalar(u8, line, '\n');
                        while (split.next()) |l| {
                            var spliit = mem.splitScalar(u8, l, '\r');
                            while (spliit.next()) |s| for (s) |c| {
                                // if (y * rect.w + x >= buffer.buf.len) break :blk;
                                buffer.cell(x, y).char = c;
                                // std.debug.print("x{d} y{d} {c}\r\n", .{ x, y, c });
                                x += 1;
                                if (x >= rect.w) {
                                    x = tx;
                                    y += 1;
                                }
                            };
                            while (x < rect.w) : (x += 1) try term.print(" ", .{});
                            y += 1;
                            x = tx;
                        }
                    }
                }
            },
            .line => {
                try term.setFg(self.getBorderFgColor());
                try term.setBg(self.getBorderBgColor());
                var x: u32 = 0;
                var y: u32 = 0;
                try term.cursorSet(rect.x, rect.y);
                try term.print("┌", .{});
                while (x < rect.w - 2) : (x += 1) try term.print("─", .{});
                try term.print("┐", .{});
                y += 1;
                while (y < rect.h - 1) : (y += 1) {
                    try term.cursorSet(rect.x, rect.y + y);
                    try term.print("│", .{});
                    try term.cursorSet(rect.x + rect.w - 1, rect.y + y);
                    try term.print("│", .{});
                }
                x = 0;
                try term.cursorSet(rect.x, rect.y + rect.h - 1);
                try term.print("└", .{});
                while (x < rect.w - 2) : (x += 1) try term.print("─", .{});
                try term.print("┘", .{});
            },
            .bg => {},
        }
        self._draw_buf = buffer;
    }
    for (self.children.items) |child| try child.draw(term);
}

fn getRect(self: *Control) Rect {
    var rect: Rect = undefined;
    if (self.parent) |parent| {
        const parent_rect = parent.getRect();
        rect.x = parent_rect.x + self.left;
        rect.y = parent_rect.y + self.top;
        rect.w = parent_rect.w + parent_rect.x - rect.x - self.right;
        rect.h = parent_rect.h + parent_rect.y - rect.y - self.bottom;
    } else return self.rect.?;
    return rect;
}
const Rect = struct { x: u32, y: u32, w: u32, h: u32 };

fn getBorderFgColor(self: *Control) Color {
    if (self.parent) |parent| {
        const parent_color = parent.getBorderFgColor();
        return parent_color.blend(&self.border_fg);
    } else return self.border_fg;
}

fn getBorderBgColor(self: *Control) Color {
    if (self.parent) |parent| {
        const parent_color = parent.getBorderBgColor();
        return parent_color.blend(&self.border_bg);
    } else return self.border_bg;
}
